"""MOSFiT: Modular light curve fitting software."""
from . import constants
from . import fitter
from . import model
from . import plotting
from . import printer
from . import utils

__version__ = '0.4.7'
__author__ = 'James Guillochon & Matt Nicholl'
__license__ = 'MIT'
